package ch.hepia.tpscala.result


sealed trait Result[+E,+A] {

  def map[B]( f: A=>B ): Result[E,B] = this match {
    case OK(value) => OK(f(value))
    case Err(value) => Err(value)
  }

  //F >: E, signifie "F est un supertype de E"
  def flatMap[B, F >:E]( f: A=>Result[F,B] ): Result[F,B] = this match {
    case OK(value) => f(value)
    case Err(value) => Err(value)
  }

  //Retourne le résultat si OK
  //transforme l'erreur en une valeur valide sinon
  //B >: A signifie "B est un supertype de A
  //Une seul implémentation possible !
  def get[B >: A]( f: E => B ): B = this match {
    case OK(value) => value
    case Err(value) => f(value)
  }

  //Une seul implémentation possible !
  def fold[B]( f: E => B, g: A => B ): B = this match {
    case OK(value) => g(value)
    case Err(value) => f(value)
  }

  def toOption: Option[A] =
    fold[Option[A]] (_ => None, Some.apply)

}
case class OK[A]( value: A ) extends Result[Nothing,A]
case class Err[E]( value: E ) extends Result[E,Nothing]
object Result {
  //Si il y a au moins une erreur dans la liste, le résultat est une erreur
  //Si il n'y a que des succès, liste les succès
  def sequence[E,A]( res: List[Result[E,A]] ): Result[E,List[A]] = {
    def reqSequence[E, A](req: List[Result[E, A]], acc: Result[E, List[A]]): Result[E, List[A]] = req match {
      case Nil => acc.map(_.reverse)
      case h :: tail => reqSequence(tail, acc.flatMap( r => h.map( a => a :: r)))
    }
    reqSequence(res, OK(Nil))
  }
}


case class Config( hostname: String, port: Int )
object Config {
  private def line2keyValue( line: String ): Result[String,(String,String)] = {
    val elements = line.split("=")
    if( elements.size != 2 ) Err("Syntax error: " + line )
    else OK( (elements(0), elements(1) ) )
  }

  private def lines2map( lines: List[String] ): Result[String,Map[String,String]] = 
    Result.sequence( lines.map(line2keyValue) ).map( _.toMap )

  private def mget[K,V](mkv: Map[K,V], key: K): Result[String, V] = mkv.get(key) match {
    case Some(v) => OK(v)
    case None => Err("Missing Key: "+key)
  }

  private def safeInt(str: String): Result[String, Int] =
    try {
      OK(str.toInt)
    }catch {
      case _ => Err("Cannot convert String :" + str + "  : to Int")
    }

  //Utilisez lines2map pour parvenir au résultat
  //Attention au clés manquantes
  def parse( lines: List[String] ): Result[String,Config] =
      for {
        keyVal <- lines2map(lines)
        h <- mget(keyVal, "hostname")
        pStr <- mget(keyVal, "port")
        p <- safeInt(pStr)
      } yield Config(h, p)
}
